/*----------------------------------------------------------------------------*
 *                                                                            *
 *   HPF Array Combining Scatter Function                                     *
 *                                                                            *
 *     - Author: Seungjo Bae & Khaled A. Alsabti                              *
 *                                                                            *
 *----------------------------------------------------------------------------*/

#include "raw_global.h"

/******************************************************************************
 *                                                                            *
 *  Function : send_sendl()
 *                                                                            *
 ******************************************************************************/

void send_sendl_raw(		/* P != power of 2 */
     int *sendl, 
     int *recvl)
{


  int spn,dpn;
  int i;
  int recv;
  int size;
#if CM5
#else
  MPI_Status status;
#endif

  size = sizeof(int);

#if CM5

#else
  MPI_Barrier(WORLD);

#endif

  for (i=1; i<tot_nproc; ++i) {

    dpn =  (my_pnum + i) % tot_nproc;
    spn =  (my_pnum - i + tot_nproc) % tot_nproc;

#if CM5

    CMMD_send_and_receive(spn,
			  CMMD_DEFAULT_TAG,
			  &recvl[spn],
			  size,
			  dpn,
			  CMMD_DEFAULT_TAG,
			  &sendl[dpn],
			  size);

#else

    MPI_Sendrecv(&sendl[dpn],
		 1,
		 MPI_INT,
		 dpn,
		 TAG,
		 &recvl[spn],
		 1,
		 MPI_INT,
		 spn,
		 TAG,
		 WORLD,
		 &status);

#endif

  }

  recvl[my_pnum] = sendl[my_pnum];

}


/******************************************************************************
 *                                                                            *
 *  Function : Preprocessing_comm
 *                                                                            *
 ******************************************************************************/


void Preprocessing_comm(  
  int sendl[],
  int recvl[],
  int send_offset[],
  int recv_offset[],
  int *total_send,
  int *total_recv)
{

  int i,tot_s,tot_r;

  send_sendl_raw(sendl, recvl);

  /*-- Calculating the send_offset and recv_offset and the send buffer  --*/
  send_offset[0] = recv_offset[0] = 0;
  tot_r = recvl[0];
  tot_s = sendl[0];
  for (i=1 ; i < tot_nproc ; i++)  {
    send_offset[i] = send_offset[i-1] + sendl[i-1];
    recv_offset[i] = recv_offset[i-1] + recvl[i-1];
    tot_r += recvl[i];
    tot_s += sendl[i];
  }
 
  *total_recv = tot_r;
  *total_send = tot_s;

/***
print_comm_matrix(sendl, 'D', *total_send, "TSC");
***/

}


/******************************************************************************
 *                                                                            *
 *  Function : LP_M_to_M_PC()
 *                                                                            *
 ******************************************************************************/

void LP_M_to_M_PC(		/* P != power of 2 */
  void *send_buf,     
  int  *sendl,   
  int  *send_offset,     
  int  datatypesize,   /* the size, in bytes, of the elements */
  void *recv_buf,  
  int  *recvl,
  int  *recv_offset)
{
  int i,spn,dpn;
  char *ptr_send,*ptr_recv;
#if CM5
#else
  MPI_Status status;
  MPI_Request request;
#endif

  ptr_send = (char *) send_buf;
  ptr_recv = (char *) recv_buf;

  for (i=1; i < tot_nproc ; ++i)  {

    dpn =  (my_pnum + i) % tot_nproc;
    spn =  (my_pnum - i + tot_nproc) % tot_nproc;

#if CM5

    if ( recvl[spn] > 0 ) 
      CMMD_free_mcb(
	CMMD_receive_async(spn,
			   CMMD_DEFAULT_TAG,
			   ptr_recv + recv_offset[spn]*datatypesize,
			   recvl[spn]*datatypesize,
			   0,
			   0)
		    );

    if ( sendl[dpn] > 0 ) 
      CMMD_send_block(dpn,
		      CMMD_DEFAULT_TAG,
		      ptr_send + send_offset[dpn]*datatypesize,
		      sendl[dpn]*datatypesize);

    while ( !CMMD_all_msgs_done() );

    CMMD_sync_with_nodes();

#else

    if ( recvl[spn] > 0 ) 

      MPI_Irecv(ptr_recv + recv_offset[spn]*datatypesize,
		recvl[spn]*datatypesize,
		MPI_BYTE,
		spn,
		TAG,
		WORLD,
		&request);

    if ( sendl[dpn] > 0 ) 

      MPI_Send(ptr_send + send_offset[dpn]*datatypesize,
	       sendl[dpn]*datatypesize,
	       MPI_BYTE,
	       dpn,
	       TAG,
	       WORLD);

    if ( recvl[spn] > 0 ) 
      MPI_Wait(&request,
	       &status);

    MPI_Barrier(WORLD);

#endif

 }

}


/******************************************************************************
 *                                                                            *
 *  Function : US_M_to_M_PC()
 *                                                                            *
 ******************************************************************************/

void US_M_to_M_PC(		/* P != power of 2 */
  void *send_buf,     
  int  *sendl,   
  int  *send_offset,     
  int  datatypesize,   /* the size, in bytes, of the elements */
  void *recv_buf,  
  int  *recvl,
  int  *recv_offset)
{
  int i, sp, tp;
  char *ptr_send,*ptr_recv;

#if CM5
#else
  MPI_Status *status;
  MPI_Request *request;
  int count;
#endif


  ptr_send = (char *) send_buf;
  ptr_recv = (char *) recv_buf;

#if CM5

  for (tp=0; tp<tot_nproc; ++tp) 
    if ( sendl[tp] && tp != my_pnum )
      CMMD_free_mcb(
	CMMD_send_async(tp,
			CMMD_DEFAULT_TAG,
			ptr_send + send_offset[tp]*datatypesize,
			sendl[tp]*datatypesize,
			0,
			0)
                   );

  for (sp=0; sp<tot_nproc; ++sp) 
    if ( recvl[sp] && sp != my_pnum )
      CMMD_free_mcb(
	CMMD_receive_async(sp,
			   CMMD_DEFAULT_TAG,
			   ptr_recv + recv_offset[sp]*datatypesize,
			   recvl[sp]*datatypesize,
			   0,
			   0)
		    );

  while ( !CMMD_all_msgs_done() );

  CMMD_sync_with_nodes();

#else

  status = (MPI_Status *) malloc(2*tot_nproc*sizeof(MPI_Status));
  request = (MPI_Request *) malloc(2*tot_nproc*sizeof(MPI_Request));
  count = 0;
  if (request == NULL) {
    printf(" Out of Memory: ....... in US_M_to_M_PC()\n"); 
    exit(0); 
  }	

  for (tp=0; tp<tot_nproc; ++tp) 
    if ( sendl[tp] && tp != my_pnum ) 
      MPI_Isend(ptr_send + send_offset[tp]*datatypesize,
	       sendl[tp]*datatypesize,
	       MPI_BYTE,
	       tp,
	       TAG,
	       WORLD,
	       &request[count++]);

  for (sp=0; sp<tot_nproc; ++sp) 
    if ( recvl[sp] && sp != my_pnum )
      MPI_Irecv(ptr_recv + recv_offset[sp]*datatypesize,
		recvl[sp]*datatypesize,
		MPI_BYTE,
		sp,
		TAG,
		WORLD,
		&request[count++]);
  
  MPI_Waitall(count, request, status);

  MPI_Barrier(WORLD);

#endif

}


/******************************************************************************
 *                                                                            *
 *  Function : M_to_M_PC()
 *                                                                            *
 ******************************************************************************/

void M_to_M_PC(
  void *send_buf,     
  int  *sendl,   
  int  *send_offset,     
  int  datatypesize,   /* the size, in bytes, of the elements */
  void *recv_buf,  
  int  *recvl,
  int  *recv_offset,  
  int algorithm)
{
  char *ptr_send,*ptr_recv;
  int inds,indr,total;

  /* Perform the local copying */
  inds = send_offset[my_pnum];
  indr = recv_offset[my_pnum]; 
   
  /* Problem : pointer of type void, how to do the copying */
  ptr_send = (char *) send_buf;
  ptr_recv = (char *) recv_buf;

  total = sendl[my_pnum]*datatypesize; 

  if (total > 0) {
    ptr_send += send_offset[my_pnum]*datatypesize;
    ptr_recv += recv_offset[my_pnum]*datatypesize;
    memcpy(ptr_recv,ptr_send,total); 
  }

  /* many-to-many PC */
  switch (algorithm) {
    case 0  : LP_M_to_M_PC(send_buf,sendl,send_offset,datatypesize,
			   recv_buf,recvl,recv_offset);
              break;
    case 1  : US_M_to_M_PC(send_buf,sendl,send_offset,datatypesize,
			   recv_buf,recvl,recv_offset);
              break;
    case 2  : 
    case 3  : 
    default : LP_M_to_M_PC(send_buf,sendl,send_offset,datatypesize,
			   recv_buf,recvl,recv_offset);
  }

}


